package yaes.framework.entity.control.speedcontrol;

import java.util.Random;

import yaes.framework.entity.control.ControlException;
import yaes.framework.entity.control.ISpeedControl;
import yaes.framework.entity.control.ISpeedControllable;
import yaes.framework.world.map.IField;
import yaes.framework.world.map.Location;
import yaes.framework.world.map.Speed;

/**
 * This model is the random waypoint model from paper [FIXME] The entity picks a
 * new, random point where to go, and it goes towards it with a constant speed.
 * When it reaches the point, it picks a new random waypoint, and moves towards
 * it.
 * 
 * The maxX and maxY values specify the maximum distance of the next waypoint
 * from the current point.
 * 
 * The waypoints are always inside the current field (if the randomly picked
 * waypoint is outside the field, a new waypoint is chosen).
 * 
 * @author Lotzi Boloni
 * 
 */
public class WaypointModel implements ISpeedControl {
	private static Random random = new Random();
	private static final double epsilon = 10; // error in reaching the
	// waypoint
	private static final double absoluteSpeed = 10; // error in reaching the
	// waypoint
	private ISpeedControllable controllable;
	private final int maxX;
	private final int maxY;
	private final IField field;
	private Location nextWaypoint;

	public WaypointModel(IField field) {
		this.field = field;
		// maximum movement 1/3 rd of the field
		maxX = (int) ((field.getXHigh() - field.getXLow()) / 3);
		maxY = (int) ((field.getYHigh() - field.getYLow()) / 3);
	}

	/**
	 * Choosing the new waypoint at a random distance of at most maxX, maxY. If
	 * the chosen waypoint is outside the field, repeat the random choice.
	 * 
	 * @param current
	 * @return
	 */
	private Location chooseNewWayPoint(Location current) {
		while (true) {
			final double nx = current.getX()
					+ WaypointModel.random.nextInt(maxX) - maxX / 2;
			if ((nx < field.getXLow()) || (nx >= field.getXHigh())) {
				continue;
			}
			final double ny = current.getY()
					+ WaypointModel.random.nextInt(maxY) - maxY / 2;
			if ((ny < field.getYLow()) || (ny >= field.getYHigh())) {
				continue;
			}
			final Location wp = new Location(nx, ny);
			return wp;
		}
	}

	/**
	 * Controls the agent
	 * 
	 * @param time
	 * @throws ControlException
	 */
	public void control(double time) throws ControlException {
		final Location current = controllable.getLocation();
		if ((nextWaypoint == null)
				|| (current.distanceTo(nextWaypoint) < WaypointModel.epsilon)) {
			nextWaypoint = chooseNewWayPoint(current);
		}
		// moving towards the waypoint
		final double dist = current.distanceTo(nextWaypoint);
		if (dist == 0) {
			return;
		}
		final double dx = nextWaypoint.getX() - current.getX();
		final double dy = nextWaypoint.getY() - current.getY();
		final Speed speed = new Speed(dx * WaypointModel.absoluteSpeed / dist,
				dy * WaypointModel.absoluteSpeed / dist);
		controllable.controlSpeed(speed);
	}

	public void setControllable(ISpeedControllable controllable) {
		this.controllable = controllable;
		controllable.addControl(this);
	}
}
