package yaes.framework.simulation;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Set;

import yaes.ui.text.TextUi;

/**
 * @author Lotzi Boloni
 * 
 * This class contains the output of a simulation.
 * 
 */
public class SimulationOutput extends Observable implements Serializable {
	static final long serialVersionUID = 1134568210844627276L;

	/**
	 * Creates a SimulationOutput object from a file
	 * 
	 * @param fileName
	 * @return
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public static SimulationOutput restore(File file)
			throws FileNotFoundException, IOException, ClassNotFoundException {
		SimulationOutput simOut = null;
		final ObjectInputStream in = new ObjectInputStream(
				new BufferedInputStream(new FileInputStream(file)));
		simOut = (SimulationOutput) in.readObject();
		in.close();
		return simOut;
	}

	/**
	 * Creates a list of simulationOutput objects from a file
	 * 
	 * @param fileName
	 * @return
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	@SuppressWarnings("unchecked")
	public static List<SimulationOutput> restoreList(File file)
			throws FileNotFoundException, IOException, ClassNotFoundException {
		List<SimulationOutput> simOut = null;
		final ObjectInputStream in = new ObjectInputStream(
				new BufferedInputStream(new FileInputStream(file)));
		simOut = (List<SimulationOutput>) in.readObject();
		in.close();
		return simOut;
	}

	/**
	 * Stores this list of SimulationOutputs in a file
	 * 
	 * @param list
	 * @param fileName
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public static void storeList(List<SimulationOutput> list, File file)
			throws FileNotFoundException, IOException {
		final ObjectOutputStream out = new ObjectOutputStream(
				new BufferedOutputStream(new FileOutputStream(file)));
		out.writeObject(list);
		out.flush();
		out.close();
	}

	private final HashMap<String, RandomVariable> variables = new HashMap<String, RandomVariable>();

	private transient IContext theContext;

	/**
	 * Creates an empty SimulationOutput.
	 */
	public SimulationOutput() {
	}

	/**
	 * Creates the SimulationOutput based on the variables from a
	 * SimulationInput
	 * 
	 * @param sim
	 * @param context
	 */
	public SimulationOutput(SimulationInput sim, IContext context) {
		if (sim == null) { // this is the case of the integratedSOP
			return;
		}
		// copying the input into the output
		for (final Iterator<String> it = sim.getParameterIterator(); it
				.hasNext();) {
			final String key = it.next();
			final SimulationInputParameter sip = sim.getParameter(key);
			switch (sip.getType()) {
			case T_DOUBLE:
				update(key, sip.getDouble());
				break;
			case T_INT:
				update(key, sip.getInt());
				break;
			case T_ENUM:
			case T_STRING:
				// FIXME: these are not yet transferred.
				continue;
			}
		}
		this.theContext = context;
	}

	/**
	 * Creates a random variable
	 * 
	 * @param name
	 * @return
	 */
	private RandomVariable createVariable(String name) {
		final RandomVariable var = new RandomVariable(name);
		variables.put(name, var);
		return var;
	}

	/**
	 * 
	 * 
	 */
	// public void setProcessParameter(String name, SimulationInputParameter
	// param) {
	// parameters.put(name, param);
	// }
	/**
	 * 
	 * 
	 */
	// public SimulationInputParameter getProcessParameter(String name) {
	// return parameters.get(name);
	// }
	/**
	 * Returns an iterator to the parameter names
	 * 
	 * @return
	 */
	public Iterator<String> getParameterIterator() {
		return variables.keySet().iterator();
	}

	/**
	 * Returns the random variable with the specified name
	 * 
	 * @param name
	 * @return
	 */
	public RandomVariable getRandomVar(String name) {
		final RandomVariable var = variables.get(name);
		if (var == null) {
			throw new Error("Random variable:" + name + " not found!");
		}
		return var;
	}

	// private HashMap<String, SimulationInputParameter> parameters = new
	// HashMap<String, SimulationInputParameter>();
	/**
	 * @return Returns the theContext.
	 */
	public IContext getTheContext() {
		return theContext;
	}

	/**
	 * Gets the value of a certain type from a randomvariable.
	 * 
	 * @param name
	 * @param type
	 * @return
	 */
	public double getValue(String name, RandomVariable.Probe type) {
		final RandomVariable var = variables.get(name);
		if (var == null) {
			// throw new Error("Can not find variable:" + name);
			TextUi.println("This variable was accessed without being set:"
					+ name + " Asummed 0.");
			update(name, 0);
			return 0.0;
		}
		return var.getValue(type);
	}

	/**
	 * Returns a set of variable names
	 */
	public Set<String> getVariableNames() {
		return variables.keySet();
	}

	/**
	 * Stores this SimulationOutput in a file
	 * 
	 * @param fileName
	 * @throws IOException
	 * @throws FileNotFoundException
	 */
	public void store(String file) throws FileNotFoundException, IOException {
		final ObjectOutputStream out = new ObjectOutputStream(
				new BufferedOutputStream(new FileOutputStream(file)));
		out.writeObject(this);
		out.flush();
		out.close();
	}

	@Override
	public String toString() {
		final StringBuffer buffer = new StringBuffer("");
		final List<String> vars = new ArrayList<String>(variables.keySet());
		Collections.sort(vars);
		for (final String key : vars) {
			final RandomVariable rv = getRandomVar(key);
			buffer.append("\t" + rv.toString());
		}
		return buffer.toString();
	}

	/**
	 * Updates a random variable
	 * 
	 * @param name
	 * @param value
	 */
	public void update(String name, double value) {
		RandomVariable var = variables.get(name);
		if (var == null) {
			var = createVariable(name);
		}
		var.update(value);
		notifyObservers();
	}
}
