package yaes.ui.propertysheet;

import java.util.Collections;
import java.util.Comparator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.List;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;

/**
 * Table model with properties (<code>Property.Property</code>) as columns
 * and nodes (<code>Property</code>) as rows. It is used as model for
 * displaying node properties in table. Each column is represented by
 * <code>Property.Property</code> object. Each row is represented by
 * <code>Property</code> object. Each cell contains
 * <code>Property.Property</code> property which equals with column object and
 * should be in property sets of row representant (<code>Property</code>).
 */
public class PropertyTableModel extends AbstractTableModel {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	static final int PROPERTY_NAMES = 0;
	static final int PROPERTY_VALUES = 1;
	/** columns of model */
	protected Vector propertyRows = null;
	protected Map properties = null;
	protected Set readOnly = null;
	/**
	 * The class of the property editor or <CODE>null</CODE> if default
	 * property editor should be used.
	 */
	@SuppressWarnings("unchecked")
	private final Comparator SORTER_TYPE = new Comparator() {
		public int compare(Object o1, Object o2) {
			String s1 = propertyValue((String) o1).getClass().getName();
			String s2 = propertyValue((String) o2).getClass().getName();
			int s = s1.compareToIgnoreCase(s2);
			if (s != 0) {
				return s;
			}
			s1 = (String) o1;
			s2 = (String) o2;
			return s1.compareToIgnoreCase(s2);
		}
	};
	/** Comparator which compares PropertyDeatils names */
	private final Comparator<String> SORTER_NAME = new Comparator<String>() {
		public int compare(String s1, String s2) {
			return String.CASE_INSENSITIVE_ORDER.compare(s1, s2);
		}
	};
	private final TableModelListener tml = new TableModelListener() {
		public void tableChanged(TableModelEvent evt) {
			PropertyTableModel model = (PropertyTableModel) evt.getSource();
			int row = evt.getFirstRow();
			if (row == -1) {
				return;
			}
			int column = evt.getColumn();
			if (column == -1) {
				model.fireTableRowsUpdated(row, row);
			} else {
				model.fireTableCellUpdated(row, column);
			}
		}
	};

	public PropertyTableModel(Map map) {
		properties = map;
		propertyRows = new Vector(properties.keySet());
	}

	public PropertyTableModel(Map map, List<?> order) {
		properties = map;
		propertyRows = new Vector(order);
	}

	public PropertyTableModel(Map map, List<?> order, Set readOnly) {
		properties = map;
		propertyRows = new Vector(order);
		this.readOnly = readOnly;
	}

	public String format(String propName) {
		return propName;
	}

	/**
	 * Getter for column count.
	 * 
	 * @return column count
	 */
	public int getColumnCount() {
		return 2;
	}

	//
	// TableModel methods
	//
	/**
	 * Getter for row count.
	 * 
	 * @return row count
	 */
	public int getRowCount() {
		return propertyRows.size();
	}

	public String[] getTags(String propName) {
		return null;
	}

	/**
	 * Getter for property.
	 * 
	 * @param row
	 *            table row index
	 * @param column
	 *            table column index
	 * @return property at (row, column)
	 */
	public Object getValueAt(int row, int column) {
		final String propertyName = (String) propertyRows.get(row);
		if (column == PropertyTableModel.PROPERTY_NAMES) {
			return propertyName;
		} else if (column == PropertyTableModel.PROPERTY_VALUES) {
			return properties.get(propertyName);
		}
		return null;
	}

	/**
	 * Cell is editable only if it has non null value.
	 * 
	 * @param row
	 *            table row index
	 * @param column
	 *            table column index
	 * @return true if cell contains non null property value
	 */
	@Override
	public boolean isCellEditable(int row, int column) {
		if (column == PropertyTableModel.PROPERTY_NAMES) {
			return false;
		}
		final String propertyName = (String) propertyRows.get(row);
		return !isReadOnly(propertyName) && (getValueAt(row, column) != null);
	}

	public boolean isReadOnly(String propertyName) {
		boolean isReadOnly = false;
		if (readOnly != null) {
			if (readOnly.contains(propertyName)) {
				isReadOnly = true;
			}
		}
		return isReadOnly;
	}

	/**
	 * Helper method to ask for a property representant of row.
	 */
	public String propertyForRow(int row) {
		return (String) propertyRows.get(row);
	}

	/**
	 * Helper method to ask for a property value corresponding to a property
	 * name.
	 */
	public Object propertyValue(String propertyName) {
		return properties.get(propertyName);
	}

	public void setOrder(List<?> order) {
		propertyRows = new Vector(order);
		fireTableDataChanged();
	}

	/**
	 * Set properties.
	 * 
	 * @param map
	 *            table of properties(name/value pairs)
	 */
	public void setProperties(Map map) {
		properties = map;
		fireTableDataChanged();
	}

	public void setPropertyValue(String propertyName, Object value) {
		properties.put(propertyName, value);
	}

	public void setReadOnly(Set readOnly) {
		this.readOnly = readOnly;
	}

	/**
	 * Setter for property.
	 * 
	 * @param value
	 *            new property value
	 * @param row
	 *            table row index
	 * @param column
	 *            table column index
	 * @return property at (row, column)
	 */
	@Override
	public void setValueAt(Object value, int row, int column) {
		final String propertyName = (String) propertyRows.get(row);
		if (value != null) {
			setPropertyValue(propertyName, value);
		}
	}

	public void sortByName() {
		Collections.sort(propertyRows, SORTER_NAME);
		fireTableDataChanged();
	}

	public void sortByType() {
		Collections.sort(propertyRows, SORTER_TYPE);
		fireTableDataChanged();
	}
}
