package yaes.ui.propertysheet;

import java.awt.Component;
import java.awt.ContainerOrderFocusTraversalPolicy;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.SystemColor;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.reflect.InvocationTargetException;
import java.util.EventObject;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.AbstractCellEditor;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.MatteBorder;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

/**
 * Table view of node properties. Table header displays property display names
 * and each cell contains (<code>PropertyPanel</code>) for
 * displaying/editting of properties. Each property row belongs to one node.
 */
public class TableSheet extends JScrollPane {
	class EditPropertyAction extends AbstractAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public void actionPerformed(ActionEvent e) {
			TableSheet.activateEditorFlag = true;
		}
	}

	/** Table cell renderer component. Paints focus border on property panel. */
	//
	// -------------- class FocusedPropertyPanel
	//
	private static class FocusedPropertyPanel extends PropertyPanel {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		boolean focused;

		public FocusedPropertyPanel(PropertyModel model, int preferences) {
			super(model, preferences);
		}

		@Override
		public void paint(Graphics g) {
			super.paint(g);
			if (focused) {
				g.setColor(UIManager.getColor("Button.focus")); // NOI18N
				g.drawRect(2, 1, getWidth() - 5, getHeight() - 4);
			}
		}

		public void setFocused(boolean focused) {
			this.focused = focused;
		}
	}

	//
	// -------------- class FocusHackedPropertyPanel
	//
	/** Table cell editor component. Contains special focus control. */
	private static class FocusHackedPropertyPanel extends PropertyPanel {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public FocusHackedPropertyPanel(PropertyModel model, int preferences) {
			super(model, preferences);
		}

		/**
		 * JTable.processKeyBinding forgets to call request focus on cell
		 * editor. So lets do it ourselves.
		 */
		@Override
		public void addNotify() {
			super.addNotify();
			requestFocus();
		}

		@Override
		public boolean isFocusable() {
			return false;
		}

		/** Forward focus request to component that can be focused. */
		@Override
		public void requestFocus() {
			final ContainerOrderFocusTraversalPolicy policy = new ContainerOrderFocusTraversalPolicy();
			policy.getDefaultComponent(this).requestFocus();
		}
	}

	//
	// -------------- class NullPanel
	//
	private static class NullPanel extends JPanel {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
	}

	/**
	 * TableCellEditor/Renderer implementation. Component returned is the
	 * PropertyPanel
	 */
	private class PropertyPanelTableCell extends AbstractCellEditor implements
			TableModelListener, TableCellEditor, TableCellRenderer {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		/** Property model used in cell editor property panel */
		private WrapperPropertyModel propModel;

		//
		// Renderer
		//
		/** Default header renderer */
		private final TableCellRenderer headerRenderer = table.getTableHeader()
				.getDefaultRenderer();

		/** Null panel is used if cell value is null */
		private NullPanel nullPanel;

		private final Border nullPanelBorder = new CompoundBorder(
				new MatteBorder(0, 0, 1, 1, SystemColor.controlDkShadow),
				new MatteBorder(1, 1, 0, 0, SystemColor.controlLtHighlight));

		private final Border nullPanelFocusBorder = new CompoundBorder(
				new CompoundBorder(nullPanelBorder, new EmptyBorder(1, 1, 1, 1)),
				new MatteBorder(1, 1, 1, 1, UIManager.getColor("Button.focus")) // NOI18N
		);

		/**
		 * Detaches listeners. Calls <code>fireEditingCanceled</code>.
		 */
		@Override
		public void cancelCellEditing() {
			// if (prop != null)
			// detachEditor();
			// super.cancelCellEditing();
		}

		/**
		 * Removes listeners and frees resources.
		 */
		private void detachEditor() {
			tableModel.removeTableModelListener(this);
			propModel = null;
		}

		/**
		 * Returns <code>null<code>.
		 * @return <code>null</code>
		 */
		public Object getCellEditorValue() {
			return null;
		}

		/**
		 * Returns editor of property.
		 * 
		 * @param table
		 * @param value
		 * @param isSelected
		 * @param r
		 *            row
		 * @param c
		 *            column
		 * @return <code>PropertyPanel</code>
		 */
		public Component getTableCellEditorComponent(JTable table,
				Object value, boolean isSelected, int r, int c) {
			final String propertyName = tableModel.propertyForRow(r);
			propModel = new WrapperPropertyModel(tableModel, propertyName, c);
			propModel.setChangeListeners(changeListeners);
			tableModel.addTableModelListener(this);
			// create property panel
			final PropertyPanel propPanel = new FocusHackedPropertyPanel(
					propModel,
					propModel.isValueReadOnly() ? PropertyPanel.PREF_READ_ONLY
							: 0);
			return propPanel;
		}

		/**
		 * Getter for actual cell renderer.
		 * 
		 * @param table
		 * @param value
		 * @param isSelected
		 * @param hasFocus
		 * @param row
		 * @param column
		 * @return <code>PropertyPanel</code>
		 */
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {
			// Header renderer
			if (row == -1) {
				final JPanel comp = new JPanel();
				return comp;
			}
			final String propertyName = tableModel.propertyForRow(row);
			propModel = new WrapperPropertyModel(tableModel, propertyName,
					column);
			propModel.setChangeListeners(changeListeners);
			if (propertyName != null) {
				final FocusedPropertyPanel propPanel = new FocusedPropertyPanel(
						propModel,
						propModel.isValueReadOnly() ? PropertyPanel.PREF_READ_ONLY
								: 0);
				propPanel.setFocused(hasFocus);
				// propPanel.setToolTipText(prop.getShortDescription());
				return propPanel;
			}
			// change it!
			if (nullPanel == null) {
				nullPanel = new NullPanel();
				nullPanel.setOpaque(true);
			} else {
				// nullPanel.setNode(node);
			}
			if (hasFocus) {
				nullPanel.setBorder(nullPanelFocusBorder);
			} else {
				nullPanel.setBorder(nullPanelBorder);
			}
			return nullPanel;
		}

		/**
		 * Return true.
		 * 
		 * @param e
		 *            event
		 * @return <code>true</code>
		 */
		@Override
		public boolean isCellEditable(EventObject e) {
			return true;
		}

		/**
		 * Cell should not be selected
		 * 
		 * @param ev
		 *            event
		 * @return <code>false</code>
		 */
		@Override
		public boolean shouldSelectCell(EventObject ev) {
			return true;
		}

		/**
		 * Detaches listeners. Calls <code>fireEditingStopped</code> and
		 * returns true.
		 * 
		 * @return true
		 */
		@Override
		public boolean stopCellEditing() {
			// if (prop != null)
			detachEditor();
			return super.stopCellEditing();
		}

		/**
		 * Table has changed. If underlied property was switched then cancel
		 * editing.
		 * 
		 * @param e
		 *            event
		 */
		public void tableChanged(TableModelEvent e) {
			final int r = table.getEditingRow();
			final int c = table.getEditingColumn();
			if ((r < 0) || (c < 0) || (r >= tableModel.getRowCount())
					|| (c >= tableModel.getColumnCount())) {
				// || (prop != tableModel.getValueAt(r, c)))
				cancelCellEditing();
			}
		}
	} // End of PropertyPanelTableCell

	/**
	 * Wraps individual property name/value pair with PropertyModel.
	 */
	//
	// -------------- class WrapperPropertyModel
	//
	public static class WrapperPropertyModel implements PropertyModel {
		PropertyChangeSupport support = new PropertyChangeSupport(this);
		private PropertyTableModel tableModel = null;
		private String propertyName = null;
		private int columnIndex = -1;
		public Vector changeListeners = null;

		public WrapperPropertyModel(PropertyTableModel tableModel,
				String propertyName, int columnIndex) {
			this.tableModel = tableModel;
			this.propertyName = propertyName;
			this.columnIndex = columnIndex;
		}

		/**
		 * Add listener to change of the value.
		 */
		public void addPropertyChangeListener(PropertyChangeListener l) {
			support.addPropertyChangeListener(l);
		}

		public void firePropertyChange() {
			support.firePropertyChange(PropertyModel.PROP_VALUE, null, null);
		}

		public Vector getChangeListeners() {
			return changeListeners;
		}

		/**
		 * The class of the property editor or <CODE>null</CODE> if default
		 * property editor should be used.
		 */
		@SuppressWarnings("unchecked")
		public Class getPropertyEditorClass() {
			return Editors.getEditorClass(getPropertyType());
		}

		/**
		 * The class of the property value.
		 */
		@SuppressWarnings("unchecked")
		public Class getPropertyType() {
			if (columnIndex == PropertyTableModel.PROPERTY_NAMES) {
				return propertyName.getClass();
			}
			return tableModel.propertyValue(propertyName).getClass();
		}

		public String[] getTags() {
			return tableModel.getTags(propertyName);
		}

		/**
		 * Getter for current value of a property.
		 */
		public Object getValue() throws InvocationTargetException {
			try {
				if (columnIndex == PropertyTableModel.PROPERTY_VALUES) {
					return tableModel.propertyValue(propertyName);
				} else if (columnIndex == PropertyTableModel.PROPERTY_NAMES) {
					return tableModel.format(propertyName);
				} else {
					return null;
				}
			} catch (final Exception e) {
				throw new InvocationTargetException(e);
			}
		}

		public boolean isPropertyName() {
			if (columnIndex == PropertyTableModel.PROPERTY_NAMES) {
				return true;
			} else {
				return false;
			}
		}

		public boolean isValueReadOnly() {
			if (isPropertyName()) {
				return false;
			}
			return tableModel.isReadOnly(propertyName);
		}

		/**
		 * Remove listener to change of the value.
		 */
		public void removePropertyChangeListener(PropertyChangeListener l) {
			support.removePropertyChangeListener(l);
		}

		public void setChangeListeners(Vector v) {
			changeListeners = v;
		}

		/**
		 * Setter for a value of a property.
		 * 
		 * @param v
		 *            the value
		 * @exeception InvocationTargetException
		 */
		public void setValue(Object v) throws InvocationTargetException {
			// impl later (if (property is read only) return;)
			try {
				tableModel.setPropertyValue(propertyName, v);
			} catch (final Exception e) {
				throw new InvocationTargetException(e);
			}
		}
	}

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Create table view with default table model.
	 */
	public static boolean activateEditorFlag;

	private static final char EditPropertyKeyCode = 'e';

	/** table */
	transient protected JTable table;

	/** model */
	private final transient PropertyTableModel tableModel;

	/**
	 * Create table view with default table model.
	 */
	private final Vector changeListeners = new Vector();

	public TableSheet(Map props) {
		tableModel = new PropertyTableModel(props);
		init();
	}

	public TableSheet(Map props, java.util.List order) {
		tableModel = new PropertyTableModel(props, order);
		init();
	}

	public TableSheet(Map props, java.util.List order, Set readOnly) {
		tableModel = new PropertyTableModel(props, order, readOnly);
		init();
	}

	/**
	 * Create table view with users table model.
	 */
	public TableSheet(PropertyTableModel tableModel) {
		this.tableModel = tableModel;
		init();
	}

	/**
	 * Allows to subclasses provide its own table.
	 * 
	 * @param tm
	 *            node table model
	 * @return table which will be placed into scroll pane
	 */
	JTable createTable() {
		return new JTable();
	}

	/**
	 * Sets resize mode of table.
	 * 
	 * @param mode -
	 *            One of 5 legal values:
	 * 
	 * <pre>
	 * JTable.AUTO_RESIZE_OFF,
	 *                                           JTable.AUTO_RESIZE_NEXT_COLUMN,
	 *                                           JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS,
	 *                                           JTable.AUTO_RESIZE_LAST_COLUMN,
	 *                                           JTable.AUTO_RESIZE_ALL_COLUMNS
	 * </pre>
	 */
	public final int getAutoResizeMode() {
		return table.getAutoResizeMode();
	}

	/**
	 * Gets preferred width of table column
	 * 
	 * @param index
	 *            column index
	 * @param width
	 *            preferred column width
	 */
	public final int getColumnPreferredWidth(int index) {
		return table.getColumnModel().getColumn(index).getPreferredWidth();
	}

	public int getSelectedRow() {
		return this.table.getSelectedRow();
	}

	private void init() {
		Editors.setCustomEditors();
		initializeView();
	}

	/**
	 * Allows to subclasses initialize table
	 * 
	 * @param t
	 */
	private void initializeTable() {
		table.setModel(tableModel);
		final PropertyPanelTableCell tableCell = new PropertyPanelTableCell();
		for (int i = 0; i < table.getColumnCount(); i++) {
			table.getColumnModel().getColumn(i).setCellEditor(tableCell);
			table.getColumnModel().getColumn(i).setCellRenderer(tableCell);
		}
		table.getTableHeader().setDefaultRenderer(tableCell);
		table.setShowGrid(false);
		table.setIntercellSpacing(new Dimension(0, 0));
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		table.getTableHeader().setReorderingAllowed(false);
		updateKeyBindings();
	}

	private void initializeView() {
		table = createTable();
		initializeTable();
		setViewportView(table);
		// do not care about focus
		setRequestFocusEnabled(false);
	}

	/**
	 * Sets resize mode of table.
	 * 
	 * @param mode -
	 *            One of 5 legal values:
	 * 
	 * <pre>
	 * JTable.AUTO_RESIZE_OFF,
	 *                                           JTable.AUTO_RESIZE_NEXT_COLUMN,
	 *                                           JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS,
	 *                                           JTable.AUTO_RESIZE_LAST_COLUMN,
	 *                                           JTable.AUTO_RESIZE_ALL_COLUMNS
	 * </pre>
	 */
	public final void setAutoResizeMode(int mode) {
		table.setAutoResizeMode(mode);
	}

	public void setChangeListeners(Vector listeners) {
		changeListeners.clear();
		for (final Iterator<?> it = listeners.iterator(); it.hasNext();) {
			changeListeners.add(it.next());
		}
	}

	/**
	 * Sets preferred width of table column
	 * 
	 * @param index
	 *            column index
	 * @param width
	 *            preferred column width
	 */
	public final void setColumnPreferredWidth(int index, int width) {
		table.getColumnModel().getColumn(index).setPreferredWidth(width);
		table.setPreferredScrollableViewportSize(table.getPreferredSize());
	}

	public void setOrder(java.util.List order) {
		tableModel.setOrder(order);
	}

	public void setProperties(Map props) {
		tableModel.setProperties(props);
	}

	public void setReadOnly(Set readOnly) {
		tableModel.setReadOnly(readOnly);
	}

	public void sortByName() {
		tableModel.sortByName();
	}

	public void sortByType() {
		tableModel.sortByType();
	}

	private void updateKeyBindings() {
		final InputMap inputMap = table.getInputMap();
		final ActionMap actionMap = table.getActionMap();
		final KeyStroke keyStroke = KeyStroke
				.getKeyStroke(TableSheet.EditPropertyKeyCode);
		actionMap.put("edit_property", new EditPropertyAction());
		inputMap.put(keyStroke, "edit_property");
	}
}
