package yaes.ui.visualization;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.geom.AffineTransform;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;

import yaes.framework.world.map.IMap;
import yaes.framework.world.map.IMovingObject;
import yaes.framework.world.map.Path;
import yaes.ui.text.TextUi;

/**
 * This is a panel which handles a map representation of the objects
 * 
 * @author Lotzi Boloni
 * 
 */
public class VisualMapPanel extends JPanel {
	/**
	 * 
	 */
	private static final long serialVersionUID = -8064900874778203803L;
	public static final int FOREGROUND_LAYER = 0;
	public static final int BACKGROUND_LAYER = 9;
	private final double scale = 1.0;
	private final List<List<Object>> objects;
	private final Map<Object, IPainter> painters = new HashMap<Object, IPainter>();
	VisualMap parent;
	private AffineTransform theTransform;

	public VisualMapPanel(VisualMap parent) {
		this.parent = parent;
		setPreferredSize(new Dimension(800, 800));
		setDoubleBuffered(true);
		setOpaque(true);
		objects = new ArrayList<List<Object>>();
		for (int i = VisualMapPanel.FOREGROUND_LAYER; i <= VisualMapPanel.BACKGROUND_LAYER; i++) {
			objects.add(new ArrayList<Object>());
		}
		theTransform = new AffineTransform();
		//theTransform.concatenate(AffineTransform.getScaleInstance(4.0, 4.0));
		//theTransform.concatenate(AffineTransform.getTranslateInstance(20, 20));
	}

	/**
	 * Simplified add Objects - tries to guess the type of object and add it
	 * based on that assumption.
	 * 
	 * @param o
	 */
	public synchronized void addObject(Object o) {
		if (o instanceof Path) {
			addObject(o, new PathPainter());
			return;
		}
		if (o instanceof IMovingObject) {
			addObject(o, new MovingObjectPainter());
			return;
		}
		TextUi.errorPrint("The added object " + o
				+ " has no associated painter");
		objects.get(1).add(o);
	}

	/**
	 * Adds an object with the associated painter.
	 * 
	 * The layer is inferred from the painter object.
	 * 
	 * @param o
	 * @param painter
	 */
	public synchronized void addObject(Object o, IPainter painter) {
		objects.get(painter.getLayer()).add(o);
		painters.put(o, painter);
		repaint();
	}

	public AffineTransform getTheTransform() {
		return theTransform;
	}

	@Override
	public void paint(Graphics g) {
		final Graphics2D g2 = (Graphics2D) g;
		invalidate();
		g2.clearRect(0, 0, getSize().width, getSize().height);
		paintVisualMap(g2);
	}

	private void paintMap(Graphics2D g, IMap fg) {
		final Image theImage = fg.getBufferedImage();
		final Image backImage = fg.getBackgroundImage();
		if ((theImage != null) && (backImage == null)) {
			g.drawImage(theImage, null, null);
		} else if ((theImage != null) && (backImage != null)) {
			g.drawImage(backImage, null, null);
		}
	}

	public synchronized void paintVisualMap(Graphics2D g) {
		final AffineTransform theTransform = g.getTransform();
		theTransform.scale(scale, scale);
		theTransform.translate(0, 0);
		g.setTransform(theTransform);
		for (int i = VisualMapPanel.BACKGROUND_LAYER; i >= VisualMapPanel.FOREGROUND_LAYER; i--) {
			for (final Object element : objects.get(i)) {
				final IPainter painter = painters.get(element);
				if (painter != null) {
					painter.paint(g, element, this);
					continue;
				}
				if (element instanceof IMap) {
					final IMap theMap = (IMap) element;
					paintMap(g, theMap);
					continue;
				}
				TextUi
						.errorPrint("Object "
								+ element
								+ " is added to the map but I don't know how to print it");
			}
		}
	}

	/**
	 * Removes the object from all the layers
	 * 
	 * @param o
	 */
	public synchronized void removeObject(Object o) {
		for (List<Object> objectl : objects) {
			objectl.remove(o);
		}
	}

}
